/*******************************************************************************
Copyright (c) 2012, Jonathan Hiller

This file is part of the AMF Tools suite. http://amf.wikispaces.com/
AMF Tools is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
AMF Tools is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.
See <http://www.opensource.org/licenses/lgpl-3.0.html> for license details.
*******************************************************************************/

#include "nMesh.h"
#include "XmlStream.h"

#include "nVolume.h"

nMesh::nMesh(void)
{
	Clear();
}


nMesh::~nMesh(void)
{
}

nMesh& nMesh::operator=(const nMesh& In)
{
	Vertices = In.Vertices;
	Volumes = In.Volumes;

	return *this;
}


void nMesh::Clear(void)
{
	Vertices.Clear();
	Volumes.clear();

}

bool nMesh::WriteXML(CXmlStreamWrite* pXML, std::string* pMessage, bool* pCancelFlag)
{
	if (pCancelFlag && *pCancelFlag) return false;

	pXML->OpenElement("mesh");

	//vertices...
	if (!Vertices.WriteXML(pXML, pMessage, pCancelFlag)) return false;

	//volumes
	for(std::vector<nVolume>::iterator it = Volumes.begin(); it != Volumes.end(); it++){
		if (!it->WriteXML(pXML, pMessage, pCancelFlag)) return false;
	}

	pXML->CloseElement(); //</mesh>
	return true;
}

bool nMesh::ReadXML(CXmlStreamRead* pXML, nAmf* pAmf, std::string* pMessage, bool* pCancelFlag)
{
	if (pCancelFlag && *pCancelFlag) return false;
	Clear();

	//load vertices...
	if (pXML->OpenElement("vertices") && Vertices.ReadXML(pXML, pAmf, pMessage, pCancelFlag)) pXML->CloseElement();
	else return false;

	//read as many volume tags as there are...
	nVolume tmpVol;
	while(pXML->OpenElement("volume", true)){ //<edge>
		if (!tmpVol.ReadXML(pXML, pAmf, pMessage, pCancelFlag)) return false;
		Volumes.push_back(tmpVol);
	}

	return true; //CheckValid(pAmf, !StrictLoad, pMessage);

}

bool nMesh::CheckValid(nAmf* pAmf, bool FixNode, std::string* pMessage)
{
	//check subnodes
	if (!Vertices.CheckValid(pAmf, FixNode, pMessage)) return false;
	for (int i=0; i<(int)Volumes.size(); i++) if(!Volumes[i].CheckValid(pAmf, FixNode, pMessage)) return false;


	//check this node

	//check if it contains any volumes
	if (Volumes.size() == 0){
		if (pMessage) *pMessage += "Warning: No volumes in mesh tag.\n";
	}

	//check if volume references any vertices that don't exist
	int NumVert = Vertices.VertexList.size();
	for (std::vector<nVolume>::iterator it = Volumes.begin(); it!=Volumes.end(); it++){
		std::vector<int>TriToDelete;
		int count=0;
		for (std::vector<nTriangle>::iterator jt = it->Triangles.begin(); jt!=it->Triangles.end(); jt++){
			bool DeleteTri = false;
			for (int i=0; i<3; i++){
				int ThisInd;
				switch (i){
				case 0: ThisInd = jt->v1; break;
				case 1: ThisInd = jt->v2; break;
				case 2: ThisInd = jt->v3; break;
				}
				if (ThisInd >= NumVert){ 
					if (FixNode) {
						DeleteTri = true;
						if (pMessage) *pMessage += "Warning: Triangle references non-existent vertex. Deleting triangle.\n"; 
					}
					else {
						if (pMessage) *pMessage += "Error: Triangle references non-existent vertex.\n";
						return false;
					}
				}
			}
			if (DeleteTri) TriToDelete.push_back(count); 
			count++;
		}
		for (int i=TriToDelete.size()-1; i >= 0; i--){ //go backwards so out indices stay correct...
			it->Triangles.erase(it->Triangles.begin()+TriToDelete[i]);
		}
	}


	return true;
}

//Utilities
bool nMesh::GetBounds(double* MinX, double* MaxX, double* MinY, double* MaxY, double* MinZ, double* MaxZ, int VolumeIndex) //returns false if no vertices
{
	if(Vertices.VertexList.size() == 0) return false;
	 
	if (VolumeIndex == -1){ //quicker, but only checks vertices
		std::vector<nVertex>::iterator it = Vertices.VertexList.begin();
		*MinX = *MaxX = it->GetX();
		*MinY = *MaxY = it->GetY();
		*MinZ = *MaxZ = it->GetZ();

		for (it = Vertices.VertexList.begin(); it!=Vertices.VertexList.end(); it++) {
			*MinX = std::min(*MinX, it->GetX());
			*MaxX = std::max(*MaxX, it->GetX());
			*MinY = std::min(*MinY, it->GetY());
			*MaxY = std::max(*MaxY, it->GetY());
			*MinZ = std::min(*MinZ, it->GetZ());
			*MaxZ = std::max(*MaxZ, it->GetZ());
		}
	}
	else { //specified some sort of volume index...
		int NumVolumes = (int)Volumes.size();
		if (VolumeIndex<0 || VolumeIndex>=NumVolumes) return false;
		nVolume* pThisVol = &Volumes[VolumeIndex];
		nVertex* pCurVertex;

		std::vector<nTriangle>::iterator it = pThisVol->Triangles.begin();
		pCurVertex = &Vertices.VertexList[it->v1];
		*MinX = *MaxX = pCurVertex->GetX();
		*MinY = *MaxY = pCurVertex->GetY();
		*MinZ = *MaxZ = pCurVertex->GetZ();

		for (it = pThisVol->Triangles.begin(); it!=pThisVol->Triangles.end(); it++) { //each triangle in this volume
			for (int i=0; i<3; i++){ //each of the 3 vertices...
				switch (i){
				case 0: pCurVertex = &Vertices.VertexList[it->v1]; break;
				case 1: pCurVertex = &Vertices.VertexList[it->v2]; break;
				case 2: pCurVertex = &Vertices.VertexList[it->v3]; break;
				}

				*MinX = std::min(*MinX, pCurVertex->GetX());
				*MaxX = std::max(*MaxX, pCurVertex->GetX());
				*MinY = std::min(*MinY, pCurVertex->GetY());
				*MaxY = std::max(*MaxY, pCurVertex->GetY());
				*MinZ = std::min(*MinZ, pCurVertex->GetZ());
				*MaxZ = std::max(*MaxZ, pCurVertex->GetZ());
			}
		}
	}
	return true;
}
